@@ -1,20 +1,31 @@
-/* See LICENSE file for copyright and license details. */
-
 /*
- * Visual layout:
- * .---------
- * | DIR: /mnt/path
- * |
- * |    file0
- * |    file1
- * |  > file2
- * |    file3
- * |    file4
- *      ...
- * |    filen
- * |
- * | Permission denied
- * '------
+ * BSD 2-Clause License
+ *
+ * Copyright (c) 2014-2016, Lazaros Koromilas <lostd@2f30.org>
+ * Copyright (c) 2014-2016, Dimitris Papastamos <sin@2f30.org>
+ * Copyright (c) 2016-2018, Arun Prakash Jana <engineerarun@gmail.com>
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ *
+ * * Redistributions of source code must retain the above copyright notice, this
+ *   list of conditions and the following disclaimer.
+ *
+ * * Redistributions in binary form must reproduce the above copyright notice,
+ *   this list of conditions and the following disclaimer in the documentation
+ *   and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
+ * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+ * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #ifdef __linux__
@@ -71,10 +82,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <time.h>
 #include <unistd.h>
-#include <readline/history.h>
-#include <readline/readline.h>
 #ifndef __USE_XOPEN_EXTENDED
 #define __USE_XOPEN_EXTENDED 1
 #endif
@@ -107,13 +117,18 @@ enabledbg()
 	FILE *fp = fopen("/tmp/nnn_debug", "w");
 
 	if (!fp) {
-		fprintf(stderr, "Cannot open debug file\n");
-		return -1;
+		fprintf(stderr, "debug: open failed! (1)\n");
+
+		fp = fopen("./nnn_debug", "w");
+		if (!fp) {
+			fprintf(stderr, "debug: open failed! (2)\n");
+			return -1;
+		}
 	}
 
 	DEBUG_FD = fileno(fp);
 	if (DEBUG_FD == -1) {
-		fprintf(stderr, "Cannot open debug file descriptor\n");
+		fprintf(stderr, "debug: open fd failed!\n");
 		return -1;
 	}
 
@@ -138,7 +153,7 @@ disabledbg()
 #endif /* DEBUGMODE */
 
 /* Macro definitions */
-#define VERSION "1.9"
+#define VERSION "2.0"
 #define GENERAL_INFO "License: BSD 2-Clause\nWebpage: https://github.com/jarun/nnn"
 
 #define LEN(x) (sizeof(x) / sizeof(*(x)))
@@ -148,7 +163,6 @@ disabledbg()
 #define TOUPPER(ch) \
 	(((ch) >= 'a' && (ch) <= 'z') ? ((ch) - 'a' + 'A') : (ch))
 #define MAX_CMD_LEN 5120
-#define CWD   "DIR: "
 #define CURSR " > "
 #define EMPTY "   "
 #define CURSYM(flag) (flag ? CURSR : EMPTY)
@@ -160,6 +174,8 @@ disabledbg()
 #define DESCRIPTOR_LEN 32
 #define _ALIGNMENT 0x10
 #define _ALIGNMENT_MASK 0xF
+#define SYMLINK_TO_DIR 0x1
+#define MAX_HOME_LEN 64
 
 /* Macros to define process spawn behaviour as flags */
 #define F_NONE     0x00  /* no flag set */
@@ -170,7 +186,7 @@ disabledbg()
 #define F_NORMAL   0x80  /* spawn child process in non-curses regular CLI mode */
 
 /* CRC8 macros */
-#define WIDTH  (8 * sizeof(unsigned char))
+#define WIDTH  (sizeof(unsigned char) << 3)
 #define TOPBIT (1 << (WIDTH - 1))
 #define POLYNOMIAL 0xD8  /* 11011 followed by 0's */
 #define CRC8_TABLE_LEN 256
@@ -202,6 +218,7 @@ disabledbg()
 typedef unsigned long ulong;
 typedef unsigned int uint;
 typedef unsigned char uchar;
+typedef unsigned short ushort;
 
 /* STRUCTURES */
 
@@ -212,7 +229,8 @@ typedef struct entry {
 	off_t size;
 	blkcnt_t blocks; /* number of 512B blocks allocated */
 	mode_t mode;
-	uint nlen; /* Length of file name; can be uchar (< NAME_MAX + 1) */
+	ushort nlen; /* Length of file name; can be uchar (< NAME_MAX + 1) */
+	uchar flags; /* Flags specific to the file */
 } __attribute__ ((packed, aligned(_ALIGNMENT))) *pEntry;
 
 /* Bookmark */
@@ -226,6 +244,7 @@ typedef struct {
 	ushort filtermode : 1;  /* Set to enter filter mode */
 	ushort mtimeorder : 1;  /* Set to sort by time modified */
 	ushort sizeorder  : 1;  /* Set to sort by file size */
+	ushort apparentsz : 1;  /* Set to sort by apparent size (disk usage) */
 	ushort blkorder   : 1;  /* Set to sort by blocks used (disk usage) */
 	ushort showhidden : 1;  /* Set to show hidden files */
 	ushort copymode   : 1;  /* Set when copying files */
@@ -241,7 +260,7 @@ typedef struct {
 /* GLOBALS */
 
 /* Configuration */
-static settings cfg = {0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 4};
+static settings cfg = {0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 4};
 
 static struct entry *dents;
 static char *pnamebuf, *pcopybuf;
@@ -251,15 +270,27 @@ static uint idletimeout, copybufpos, copybuflen;
 static char *player;
 static char *copier;
 static char *editor;
-static char *desktop_manager;
+static char *dmanager; /* desktop file manager */
 static blkcnt_t ent_blocks;
 static blkcnt_t dir_blocks;
 static ulong num_files;
 static uint open_max;
 static bm bookmark[BM_MAX];
-
-static uchar crc8table[CRC8_TABLE_LEN];
+static size_t g_homelen;
 static uchar g_crc;
+static uchar BLK_SHIFT = 9;
+
+/* CRC data */
+static uchar crc8table[CRC8_TABLE_LEN] __attribute__ ((aligned));
+
+/* For use in functions which are isolated and don't return the buffer */
+static char g_buf[MAX_CMD_LEN] __attribute__ ((aligned));
+
+/* Buffer for file path copy file */
+static char g_cppath[MAX_HOME_LEN] __attribute__ ((aligned));
+
+/* Buffer to store HOME path, for help and file details */
+static char g_homepath[MAX_HOME_LEN] __attribute__ ((aligned));
 
 #ifdef LINUX_INOTIFY
 static int inotify_fd, inotify_wd = -1;
@@ -316,12 +347,6 @@ static const char messages[][16] = {
 	"%F %T %z",
 };
 
-/* For use in functions which are isolated and don't return the buffer */
-static char g_buf[MAX_CMD_LEN] __attribute__ ((aligned));
-
-/* Buffer for file path copy file */
-static char g_cppath[48] __attribute__ ((aligned));
-
 /* Forward declarations */
 static void redraw(char *path);
 
@@ -385,6 +410,8 @@ printerr(int linenum)
 {
 	exitcurses();
 	fprintf(stderr, "line %d: (%d) %s\n", linenum, errno, strerror(errno));
+	if (cfg.noxdisplay && g_cppath[0])
+		unlink(g_cppath);
 	exit(1);
 }
 
@@ -600,6 +627,9 @@ xbasename(char *path)
 static void
 writecp(const char *buf, const size_t buflen)
 {
+	if (!g_cppath[0])
+		return;
+
 	FILE *fp = fopen(g_cppath, "w");
 
 	if (fp) {
@@ -712,8 +742,8 @@ spawn(const char *file, const char *arg1, const char *arg2, const char *dir, uch
 
 		/* Show a marker (to indicate nnn spawned shell) */
 		if (flag & F_MARKER && shlvl != NULL) {
-			printf("\n +-++-++-+\n | n n n |\n +-++-++-+\n\n");
-			printf("Next shell level: %d\n", atoi(shlvl) + 1);
+			fprintf(stdout, "\n +-++-++-+\n | n n n |\n +-++-++-+\n\n");
+			fprintf(stdout, "Next shell level: %d\n", atoi(shlvl) + 1);
 		}
 
 		/* Suppress stdout and stderr */
@@ -846,25 +876,6 @@ xchartohex(char c)
 	return c;
 }
 
-/* Trim all whitespace from both ends, / from end */
-static char *
-strstrip(char *s)
-{
-	if (!s || !*s)
-		return s;
-
-	size_t len = strlen(s) - 1;
-
-	while (len != 0 && (isspace(s[len]) || s[len] == '/'))
-		--len;
-	s[len + 1] = '\0';
-
-	while (*s && isspace(*s))
-		++s;
-
-	return s;
-}
-
 static char *
 getmime(const char *file)
 {
@@ -1055,7 +1066,7 @@ matches(char *fltr)
 
 	ndents = fill(&dents, visible, &re);
 	regfree(&re);
-	if (ndents == 0)
+	if (!ndents)
 		return 0;
 
 	qsort(dents, ndents, sizeof(*dents), entrycmp);
@@ -1095,9 +1106,9 @@ filterentries(char *path)
 
 			wcstombs(ln, wln, REGEX_MAX);
 			ndents = total;
-			if (matches(pln) == -1)
-				continue;
-			redraw(path);
+			if (matches(pln) != -1)
+				redraw(path);
+
 			printprompt(ln);
 			continue;
 		}
@@ -1320,9 +1331,12 @@ mkpath(char *dir, char *name, char *out, size_t n)
 }
 
 static void
-parsebmstr(char *bms)
+parsebmstr()
 {
 	int i = 0;
+	char *bms = getenv("NNN_BMS");
+	if (!bms)
+		return;
 
 	while (*bms && i < BM_MAX) {
 		bookmark[i].key = bms;
@@ -1557,12 +1571,15 @@ printent_long(struct entry *ent, int sel, uint namecols)
 
 	if (S_ISDIR(ent->mode)) {
 		if (cfg.blkorder)
-			printw("%s%-16.16s %8.8s/ %s/\n", CURSYM(sel), buf, coolsize(ent->blocks << 9), pname);
+			printw("%s%-16.16s %8.8s/ %s/\n", CURSYM(sel), buf, coolsize(ent->blocks << BLK_SHIFT), pname);
 		else
 			printw("%s%-16.16s        /  %s/\n", CURSYM(sel), buf, pname);
-	} else if (S_ISLNK(ent->mode))
-		printw("%s%-16.16s        @  %s@\n", CURSYM(sel), buf, pname);
-	else if (S_ISSOCK(ent->mode))
+	} else if (S_ISLNK(ent->mode)) {
+		if (ent->flags & SYMLINK_TO_DIR)
+			printw("%s%-16.16s       @/  %s@\n", CURSYM(sel), buf, pname);
+		else
+			printw("%s%-16.16s        @  %s@\n", CURSYM(sel), buf, pname);
+	} else if (S_ISSOCK(ent->mode))
 		printw("%s%-16.16s        =  %s=\n", CURSYM(sel), buf, pname);
 	else if (S_ISFIFO(ent->mode))
 		printw("%s%-16.16s        |  %s|\n", CURSYM(sel), buf, pname);
@@ -1572,12 +1589,12 @@ printent_long(struct entry *ent, int sel, uint namecols)
 		printw("%s%-16.16s        c  %s\n", CURSYM(sel), buf, pname);
 	else if (ent->mode & 0100) {
 		if (cfg.blkorder)
-			printw("%s%-16.16s %8.8s* %s*\n", CURSYM(sel), buf, coolsize(ent->blocks << 9), pname);
+			printw("%s%-16.16s %8.8s* %s*\n", CURSYM(sel), buf, coolsize(ent->blocks << BLK_SHIFT), pname);
 		else
 			printw("%s%-16.16s %8.8s* %s*\n", CURSYM(sel), buf, coolsize(ent->size), pname);
 	} else {
 		if (cfg.blkorder)
-			printw("%s%-16.16s %8.8s  %s\n", CURSYM(sel), buf, coolsize(ent->blocks << 9), pname);
+			printw("%s%-16.16s %8.8s  %s\n", CURSYM(sel), buf, coolsize(ent->blocks << BLK_SHIFT), pname);
 		else
 			printw("%s%-16.16s %8.8s  %s\n", CURSYM(sel), buf, coolsize(ent->size), pname);
 	}
@@ -1761,8 +1778,15 @@ show_stats(char *fpath, char *fname, struct stat *sb)
 	char desc[DESCRIPTOR_LEN];
 	char *perms = get_lsperms(sb->st_mode, desc);
 	char *p, *begin = g_buf;
-	char tmp[] = "/tmp/nnnXXXXXX";
-	int fd = mkstemp(tmp);
+
+	if (g_homepath[0])
+		xstrlcpy(g_homepath + g_homelen - 1, "/.nnnXXXXXX", MAX_HOME_LEN - g_homelen);
+	else {
+		printmsg(messages[STR_NOHOME_ID]);
+		return -1;
+	}
+
+	int fd = mkstemp(g_homepath);
 
 	if (fd == -1)
 		return -1;
@@ -1775,6 +1799,10 @@ show_stats(char *fpath, char *fname, struct stat *sb)
 		ssize_t len = readlink(fpath, g_buf, MAX_CMD_LEN);
 
 		if (len != -1) {
+			struct stat tgtsb;
+			if (!stat(fpath, &tgtsb) && S_ISDIR(tgtsb.st_mode))
+				g_buf[len++] = '/';
+
 			g_buf[len] = '\0';
 
 			/*
@@ -1849,8 +1877,8 @@ show_stats(char *fpath, char *fname, struct stat *sb)
 	close(fd);
 
 	exitcurses();
-	get_output(NULL, 0, "cat", tmp, NULL, 1);
-	unlink(tmp);
+	get_output(NULL, 0, "cat", g_homepath, NULL, 1);
+	unlink(g_homepath);
 	refresh();
 	return 0;
 }
@@ -1909,60 +1937,65 @@ handle_archive(char *fpath, char *arg, char *dir)
 static int
 show_help(char *path)
 {
-	char tmp[] = "/tmp/nnnXXXXXX";
-	int i = 0, fd = mkstemp(tmp);
+	if (g_homepath[0])
+		xstrlcpy(g_homepath + g_homelen - 1, "/.nnnXXXXXX", MAX_HOME_LEN - g_homelen);
+	else {
+		printmsg(messages[STR_NOHOME_ID]);
+		return -1;
+	}
+
+	int i = 0, fd = mkstemp(g_homepath);
 	char *start, *end;
 	static char helpstr[] = {
-	   "cKey | Function\n"
-	     "e- + -\n"
-      "7↑, k, ^P | Prev entry\n"
-      "7↓, j, ^N | Next entry\n"
-      "7PgUp, ^U | Scroll half page up\n"
-      "7PgDn, ^D | Scroll half page down\n"
-"1Home, g, ^, ^A | First entry\n"
- "2End, G, $, ^E | Last entry\n"
-   "4→, ↵, l, ^M | Open file/enter dir\n"
-"1←, Bksp, h, ^H | Parent dir\n"
-	    "d^O | Open with...\n"
-    "5Insert, ^I | Toggle nav-as-you-type\n"
-	     "e~ | Go HOME\n"
-	     "e& | Start-up dir\n"
-	     "e- | Last visited dir\n"
-	     "e/ | Filter entries\n"
-	    "d^/ | Open desktop search app\n"
-	     "e. | Toggle show . files\n"
-	    "d^B | Bookmark prompt\n"
-	     "eb | Pin current dir\n"
-	    "d^V | Go to pinned dir\n"
-	     "ec | Change dir prompt\n"
-	     "ed | Toggle detail view\n"
-	     "eD | File details\n"
-	     "em | Brief media info\n"
-	     "eM | Full media info\n"
-	     "en | Create new\n"
-	    "d^R | Rename entry\n"
-	     "er | Open dir in vidir\n"
-	     "es | Toggle sort by size\n"
-	 "aS, ^J | Toggle du mode\n"
-	     "et | Toggle sort by mtime\n"
-	 "a!, ^] | Spawn SHELL in dir\n"
-	     "eR | Run custom script\n"
-	     "ee | Edit entry in EDITOR\n"
-	     "eo | Open DE filemanager\n"
-	     "ep | Open entry in PAGER\n"
-	     "ef | Archive entry\n"
-	     "eF | List archive\n"
-	    "d^F | Extract archive\n"
-     "6Space, ^K | Copy file path\n"
-	    "d^Y | Toggle multi-copy\n"
-	    "d^T | Toggle path quote\n"
-	    "d^L | Redraw, clear prompt\n"
-#ifdef __linux__
-	     "eL | Lock terminal\n"
-#endif
-	     "e? | Help, settings\n"
-	 "aQ, ^G | Quit and cd\n"
-	 "aq, ^X | Quit\n\n"};
+	   "cKey  Desc\n"
+	     "e----\n"
+      "7↑, k, ^P  Prev entry\n"
+      "7↓, j, ^N  Next entry\n"
+      "7PgUp, ^U  Scroll half page up\n"
+      "7PgDn, ^D  Scroll half page down\n"
+"1Home, g, ^, ^A  First entry\n"
+ "2End, G, $, ^E  Last entry\n"
+   "4→, ↵, l, ^M  Open file/enter dir\n"
+"1←, Bksp, h, ^H  Parent dir\n"
+	    "d^O  Open with...\n"
+    "5Insert, ^I  Toggle nav-as-you-type\n"
+	     "e~  Go HOME\n"
+	     "e&  Start-up dir\n"
+	     "e-  Last visited dir\n"
+	     "e/  Filter entries\n"
+	    "d^/  Open desktop search app\n"
+	     "e.  Toggle show . files\n"
+	    "d^B  Bookmark prompt\n"
+	     "eb  Pin current dir\n"
+	    "d^V  Go to pinned dir\n"
+	     "ec  Change dir prompt\n"
+	     "ed  Toggle detail view\n"
+	     "eD  File details\n"
+	     "em  Brief media info\n"
+	     "eM  Full media info\n"
+	     "en  Create new\n"
+	    "d^R  Rename entry\n"
+	     "er  Open dir in vidir\n"
+	     "es  Toggle sort by size\n"
+	     "eS  Toggle apparent size\n"
+	    "d^J  Toggle du mode\n"
+	     "et  Toggle sort by mtime\n"
+	 "a!, ^]  Spawn SHELL in dir\n"
+	     "eR  Run custom script\n"
+	     "ee  Edit entry in EDITOR\n"
+	     "eo  Open DE filemanager\n"
+	     "ep  Open entry in PAGER\n"
+	     "ef  Archive entry\n"
+	     "eF  List archive\n"
+	    "d^F  Extract archive\n"
+     "6Space, ^K  Copy file path\n"
+	    "d^Y  Toggle multi-copy\n"
+	    "d^T  Toggle path quote\n"
+	    "d^L  Redraw, clear prompt\n"
+	     "eL  Lock terminal\n"
+	     "e?  Help, settings\n"
+	 "aQ, ^G  Quit and cd\n"
+	 "aq, ^X  Quit\n\n"};
 
 	if (fd == -1)
 		return -1;
@@ -1996,8 +2029,8 @@ show_help(char *path)
 
 	if (editor)
 		dprintf(fd, "NNN_USE_EDITOR: %s\n", editor);
-	if (desktop_manager)
-		dprintf(fd, "NNN_DE_FILE_MANAGER: %s\n", desktop_manager);
+	if (dmanager)
+		dprintf(fd, "NNN_DE_FILE_MANAGER: %s\n", dmanager);
 	if (idletimeout)
 		dprintf(fd, "NNN_IDLE_TIMEOUT: %d secs\n", idletimeout);
 	if (copier)
@@ -2030,12 +2063,14 @@ show_help(char *path)
 	close(fd);
 
 	exitcurses();
-	get_output(NULL, 0, "cat", tmp, NULL, 1);
-	unlink(tmp);
+	get_output(NULL, 0, "cat", g_homepath, NULL, 1);
+	unlink(g_homepath);
 	refresh();
 	return 0;
 }
 
+int (*nftw_fn) (const char *fpath, const struct stat *sb, int typeflag, struct FTW *ftwbuf);
+
 static int
 sum_bsizes(const char *fpath, const struct stat *sb,
 	   int typeflag, struct FTW *ftwbuf)
@@ -2047,6 +2082,17 @@ sum_bsizes(const char *fpath, const struct stat *sb,
 	return 0;
 }
 
+static int
+sum_sizes(const char *fpath, const struct stat *sb,
+	   int typeflag, struct FTW *ftwbuf)
+{
+	if (sb->st_size && (typeflag == FTW_F || typeflag == FTW_D))
+		ent_blocks += sb->st_size;
+
+	++num_files;
+	return 0;
+}
+
 static int
 dentfill(char *path, struct entry **dents,
 	 int (*filter)(regex_t *, char *), regex_t *re)
@@ -2099,16 +2145,14 @@ dentfill(char *path, struct entry **dents,
 					ent_blocks = 0;
 					mkpath(path, namep, g_buf, PATH_MAX);
 
-					if (nftw(g_buf, sum_bsizes, open_max, FTW_MOUNT | FTW_PHYS) == -1) {
+					if (nftw(g_buf, nftw_fn, open_max, FTW_MOUNT | FTW_PHYS) == -1) {
 						printmsg(messages[STR_NFTWFAIL_ID]);
-						dir_blocks += sb.st_blocks;
+						dir_blocks += (cfg.apparentsz ? sb.st_size : sb.st_blocks);
 					} else
 						dir_blocks += ent_blocks;
 				}
 			} else {
-				if (sb.st_blocks)
-					dir_blocks += sb.st_blocks;
-
+				dir_blocks += (cfg.apparentsz ? sb.st_size : sb.st_blocks);
 				++num_files;
 			}
 
@@ -2177,9 +2221,9 @@ dentfill(char *path, struct entry **dents,
 				num_saved = num_files + 1;
 				mkpath(path, namep, g_buf, PATH_MAX);
 
-				if (nftw(g_buf, sum_bsizes, open_max, FTW_MOUNT | FTW_PHYS) == -1) {
+				if (nftw(g_buf, nftw_fn, open_max, FTW_MOUNT | FTW_PHYS) == -1) {
 					printmsg(messages[STR_NFTWFAIL_ID]);
-					dentp->blocks = sb.st_blocks;
+					dentp->blocks = (cfg.apparentsz ? sb.st_size : sb.st_blocks);
 				} else
 					dentp->blocks = ent_blocks;
 
@@ -2188,12 +2232,21 @@ dentfill(char *path, struct entry **dents,
 				else
 					num_files = num_saved;
 			} else {
-				dentp->blocks = sb.st_blocks;
+				dentp->blocks = (cfg.apparentsz ? sb.st_size : sb.st_blocks);
 				dir_blocks += dentp->blocks;
 				++num_files;
 			}
 		}
 
+		/* Flag if this is a symlink to a dir */
+		if (S_ISLNK(sb.st_mode))
+			if (!fstatat(fd, namep, &sb, 0)) {
+				if (S_ISDIR(sb.st_mode))
+					dentp->flags |= SYMLINK_TO_DIR;
+				else
+					dentp->flags &= ~SYMLINK_TO_DIR;
+			}
+
 		++n;
 	}
 
@@ -2262,7 +2315,7 @@ populate(char *path, char *oldname, char *fltr)
 
 	ndents = dentfill(path, &dents, visible, &re);
 	regfree(&re);
-	if (ndents == 0)
+	if (!ndents)
 		return 0;
 
 	qsort(dents, ndents, sizeof(*dents), entrycmp);
@@ -2321,10 +2374,11 @@ redraw(char *path)
 	if (ncols > PATH_MAX)
 		ncols = PATH_MAX;
 
+	attron(A_UNDERLINE);
 	/* No text wrapping in cwd line */
-	/* Show CWD: - strlen(CWD) - 1 = 6 */
-	g_buf[ncols - 6] = '\0';
-	printw(CWD "%s\n\n", g_buf);
+	g_buf[ncols - 1] = '\0';
+	printw("%s\n\n", g_buf);
+	attroff(A_UNDERLINE);
 
 	/* Fallback to light mode if less than 35 columns */
 	if (ncols < 35 && cfg.showdetail) {
@@ -2382,7 +2436,14 @@ redraw(char *path)
 				snprintf(buf, NAME_MAX + 65, "%d/%d %s[%s%s]",
 					 cur + 1, ndents, sort, unescape(dents[cur].name, 0), get_file_sym(dents[cur].mode));
 			else {
-				i = snprintf(buf, 64, "%d/%d du: %s (%lu files) ", cur + 1, ndents, coolsize(dir_blocks << 9), num_files);
+				i = snprintf(buf, 64, "%d/%d ", cur + 1, ndents);
+
+				if (cfg.apparentsz)
+					buf[i++] = 'a', buf[i++] = 'p';
+				else
+					buf[i++] = 'd', buf[i++] = 'u';
+
+				i += snprintf(buf + i, 64, ": %s (%lu files) ", coolsize(dir_blocks << BLK_SHIFT), num_files);
 				snprintf(buf + i, NAME_MAX, "vol: %s free [%s%s]",
 					 coolsize(get_fs_info(path, FREE)), unescape(dents[cur].name, 0), get_file_sym(dents[cur].mode));
 			}
@@ -2409,7 +2470,7 @@ browse(char *ipath, char *ifilter)
 	static char oldname[NAME_MAX + 1] __attribute__ ((aligned));
 	char *dir, *tmp, *run = NULL, *env = NULL;
 	struct stat sb;
-	int r, fd, presel, ncp = 0, copystartid = 0, copyendid = 0;
+	int r, fd, truecd, presel, ncp = 0, copystartid = 0, copyendid = 0;
 	enum action sel = SEL_RUNARG + 1;
 	bool dir_changed = FALSE;
 
@@ -2511,7 +2572,7 @@ nochange:
 			goto begin;
 		case SEL_GOIN:
 			/* Cannot descend in empty directories */
-			if (ndents == 0)
+			if (!ndents)
 				goto begin;
 
 			mkpath(path, dents[cur].name, newpath, PATH_MAX);
@@ -2609,77 +2670,34 @@ nochange:
 			cur = ndents - 1;
 			break;
 		case SEL_CD:
-		{
-			char *input;
-			int truecd;
-
-			/* Save the program start dir */
-			tmp = getcwd(newpath, PATH_MAX);
-			if (tmp == NULL) {
-				printwarn();
-				goto nochange;
-			}
-
-			/* Switch to current path for readline(3) */
-			if (chdir(path) == -1) {
-				printwarn();
-				goto nochange;
-			}
-
-			exitcurses();
-			tmp = readline("cd: ");
-			refresh();
-
-			/* Change back to program start dir */
-			if (chdir(newpath) == -1)
-				printwarn();
-
-			if (tmp[0] == '\0')
-				break;
-
-			/* Add to readline(3) history */
-			add_history(tmp);
-
-			input = tmp;
-			tmp = strstrip(tmp);
-			if (tmp[0] == '\0') {
-				free(input);
-				break;
-			}
-
 			truecd = 0;
+			tmp = xreadline(NULL, "cd: ");
+			if (tmp == NULL || tmp[0] == '\0')
+				break;
 
 			if (tmp[0] == '~') {
 				/* Expand ~ to HOME absolute path */
-				char *home = getenv("HOME");
-
-				if (home)
-					snprintf(newpath, PATH_MAX, "%s%s", home, tmp + 1);
+				dir = getenv("HOME");
+				if (dir)
+					snprintf(newpath, PATH_MAX, "%s%s", dir, tmp + 1);
 				else {
-					free(input);
 					printmsg(messages[STR_NOHOME_ID]);
 					goto nochange;
 				}
 			} else if (tmp[0] == '-' && tmp[1] == '\0') {
-				if (lastdir[0] == '\0') {
-					free(input);
+				if (lastdir[0] == '\0')
 					break;
-				}
 
 				/* Switch to last visited dir */
 				xstrlcpy(newpath, lastdir, PATH_MAX);
 				truecd = 1;
 			} else if ((r = all_dots(tmp))) {
-				if (r == 1) {
+				if (r == 1)
 					/* Always in the current dir */
-					free(input);
 					break;
-				}
 
 				/* Show a message if already at / */
 				if (istopdir(path)) {
-					free(input);
-
 					/* Continue in navigate-as-you-type mode, if enabled */
 					if (cfg.filtermode)
 						presel = FILTER;
@@ -2701,7 +2719,6 @@ nochange:
 					dir = xdirname(dir);
 					if (access(dir, R_OK) == -1) {
 						printwarn();
-						free(input);
 						goto nochange;
 					}
 				}
@@ -2720,8 +2737,6 @@ nochange:
 			} else
 				mkpath(path, tmp, newpath, PATH_MAX);
 
-			free(input);
-
 			if (!xdiraccess(newpath))
 				goto nochange;
 
@@ -2749,7 +2764,6 @@ nochange:
 			if (cfg.filtermode)
 				presel = FILTER;
 			goto begin;
-		}
 		case SEL_CDHOME:
 			dir = getenv("HOME");
 			if (dir == NULL) {
@@ -2854,7 +2868,7 @@ nochange:
 			copyfilter();
 			DPRINTF_S(fltr);
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto nochange;
 		case SEL_MFLTR:
@@ -2863,7 +2877,7 @@ nochange:
 				presel = FILTER;
 			else {
 				/* Save current */
-				if (ndents > 0)
+				if (ndents)
 					copycurname();
 
 				/* Start watching the directory */
@@ -2882,81 +2896,99 @@ nochange:
 			cfg.showdetail ^= 1;
 			cfg.showdetail ? (printptr = &printent_long) : (printptr = &printent);
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto begin;
 		case SEL_STATS:
-			if (ndents > 0) {
-				mkpath(path, dents[cur].name, newpath, PATH_MAX);
+			if (!ndents)
+				break;
 
-				if (lstat(newpath, &sb) == -1) {
-					if (dents)
-						dentfree(dents);
-					errexit();
-				} else {
-					if (show_stats(newpath, dents[cur].name, &sb) < 0) {
-						printwarn();
-						goto nochange;
-					}
-				}
+			mkpath(path, dents[cur].name, newpath, PATH_MAX);
+
+			if (lstat(newpath, &sb) == -1) {
+				if (dents)
+					dentfree(dents);
+				errexit();
+			}
+
+			if (show_stats(newpath, dents[cur].name, &sb) < 0) {
+				printwarn();
+				goto nochange;
 			}
 			break;
 		case SEL_LIST: // fallthrough
 		case SEL_EXTRACT: // fallthrough
 		case SEL_MEDIA: // fallthrough
 		case SEL_FMEDIA:
-			if (ndents > 0) {
-				mkpath(path, dents[cur].name, newpath, PATH_MAX);
+			if (!ndents)
+				break;
 
+			mkpath(path, dents[cur].name, newpath, PATH_MAX);
+
+			if (sel == SEL_MEDIA || sel == SEL_FMEDIA)
+				r = show_mediainfo(newpath, run);
+			else
+				r = handle_archive(newpath, run, path);
+
+			if (r == -1) {
+				xstrlcpy(newpath, "missing ", PATH_MAX);
 				if (sel == SEL_MEDIA || sel == SEL_FMEDIA)
-					r = show_mediainfo(newpath, run);
+					xstrlcpy(newpath + 8, utils[cfg.metaviewer], 32);
 				else
-					r = handle_archive(newpath, run, path);
+					xstrlcpy(newpath + 8, utils[ATOOL], 32);
 
-				if (r == -1) {
-					xstrlcpy(newpath, "missing ", PATH_MAX);
-					if (sel == SEL_MEDIA || sel == SEL_FMEDIA)
-						xstrlcpy(newpath + 8, utils[cfg.metaviewer], 32);
-					else
-						xstrlcpy(newpath + 8, utils[ATOOL], 32);
+				printmsg(newpath);
+				goto nochange;
+			}
 
-					printmsg(newpath);
-					goto nochange;
-				}
+			/* In case of successful archive extract, reload contents */
+			if (sel == SEL_EXTRACT) {
+				/* Continue in navigate-as-you-type mode, if enabled */
+				if (cfg.filtermode)
+					presel = FILTER;
 
-				/* In case of successful archive extract, reload contents */
-				if (sel == SEL_EXTRACT) {
-					/* Continue in navigate-as-you-type mode, if enabled */
-					if (cfg.filtermode)
-						presel = FILTER;
+				/* Save current */
+				copycurname();
 
-					/* Save current */
-					copycurname();
-
-					/* Repopulate as directory content may have changed */
-					goto begin;
-				}
+				/* Repopulate as directory content may have changed */
+				goto begin;
 			}
 			break;
 		case SEL_DFB:
-			if (!desktop_manager) {
+			if (!dmanager) {
 				printmsg("set NNN_DE_FILE_MANAGER");
 				goto nochange;
 			}
 
-			spawn(desktop_manager, path, NULL, path, F_NOWAIT | F_NOTRACE);
+			spawn(dmanager, path, NULL, path, F_NOWAIT | F_NOTRACE);
 			break;
 		case SEL_FSIZE:
 			cfg.sizeorder ^= 1;
 			cfg.mtimeorder = 0;
+			cfg.apparentsz = 0;
 			cfg.blkorder = 0;
 			cfg.copymode = 0;
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto begin;
+		case SEL_ASIZE:
+			cfg.apparentsz ^= 1;
+			if (cfg.apparentsz) {
+				nftw_fn = &sum_sizes;
+				cfg.blkorder = 1;
+				BLK_SHIFT = 0;
+			} else
+				cfg.blkorder = 0; // fallthrough
 		case SEL_BSIZE:
-			cfg.blkorder ^= 1;
+			if (sel == SEL_BSIZE) {
+				if (!cfg.apparentsz)
+					cfg.blkorder ^= 1;
+				nftw_fn = &sum_bsizes;
+				cfg.apparentsz = 0;
+				BLK_SHIFT = 9;
+			}
+
 			if (cfg.blkorder) {
 				cfg.showdetail = 1;
 				printptr = &printent_long;
@@ -2965,21 +2997,22 @@ nochange:
 			cfg.sizeorder = 0;
 			cfg.copymode = 0;
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto begin;
 		case SEL_MTIME:
 			cfg.mtimeorder ^= 1;
 			cfg.sizeorder = 0;
+			cfg.apparentsz = 0;
 			cfg.blkorder = 0;
 			cfg.copymode = 0;
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto begin;
 		case SEL_REDRAW:
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto begin;
 		case SEL_COPY:
@@ -3079,7 +3112,7 @@ nochange:
 			goto nochange;
 		case SEL_OPEN: // fallthrough
 		case SEL_ARCHIVE:
-			if (ndents <= 0)
+			if (!ndents)
 				break; // fallthrough
 		case SEL_NEW:
 			if (sel == SEL_OPEN)
@@ -3169,7 +3202,7 @@ nochange:
 			xstrlcpy(oldname, tmp, NAME_MAX + 1);
 			goto begin;
 		case SEL_RENAME:
-			if (ndents <= 0)
+			if (!ndents)
 				break;
 
 			tmp = xreadline(dents[cur].name, "");
@@ -3225,7 +3258,7 @@ nochange:
 			spawn(utils[VIDIR], ".", NULL, path, F_NORMAL);
 
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 			goto begin;
 		case SEL_HELP:
@@ -3254,11 +3287,12 @@ nochange:
 
 					char *curfile = NULL;
 
-					if (ndents > 0)
+					if (ndents)
 						curfile = dents[cur].name;
 
 					spawn(run, tmp, curfile, path, F_NORMAL | F_SIGINT);
-				}
+				} else
+					printmsg("set NNN_SCRIPT");
 			} else {
 				spawn(run, NULL, NULL, path, F_NORMAL | F_MARKER);
 
@@ -3268,7 +3302,7 @@ nochange:
 			}
 
 			/* Save current */
-			if (ndents > 0)
+			if (ndents)
 				copycurname();
 
 			/* Repopulate as directory content may have changed */
@@ -3279,20 +3313,18 @@ nochange:
 				run = editor ? editor : xgetenv("EDITOR", "vi");
 			spawn(run, dents[cur].name, NULL, path, F_NORMAL);
 			break;
-#ifdef __linux__
 		case SEL_LOCK:
 			spawn(player, "", "screensaver", NULL, F_NORMAL | F_SIGINT);
 			break;
-#endif
 		case SEL_CDQUIT:
 		{
-			char *tmpfile = "/tmp/nnn";
-
 			tmp = getenv("NNN_TMPFILE");
-			if (tmp)
-				tmpfile = tmp;
+			if (!tmp) {
+				printmsg("set NNN_TMPFILE");
+				goto nochange;
+			}
 
-			FILE *fp = fopen(tmpfile, "w");
+			FILE *fp = fopen(tmp, "w");
 
 			if (fp) {
 				fprintf(fp, "cd \"%s\"", path);
@@ -3317,21 +3349,22 @@ nochange:
 static void
 usage(void)
 {
-	printf("usage: nnn [-b key] [-c N] [-e] [-i] [-l]\n"
+	fprintf(stdout,
+		"usage: nnn [-b key] [-c N] [-e] [-i] [-l]\n"
 		"           [-p nlay] [-S] [-v] [-h] [PATH]\n\n"
 		"The missing terminal file manager for X.\n\n"
-		"positional arguments:\n"
+		"positional args:\n"
 		"  PATH   start dir [default: current dir]\n\n"
-		"optional arguments:\n"
-		" -b key  specify bookmark key to open\n"
-		" -c N    specify dir color, disables if N>7\n"
+		"optional args:\n"
+		" -b key  bookmark key to open\n"
+		" -c N    dir color, disables if N>7\n"
 		" -e      use exiftool instead of mediainfo\n"
 		" -i      start in navigate-as-you-type mode\n"
-		" -l      start in light mode (fewer details)\n"
+		" -l      start in light mode\n"
 		" -p nlay path to custom nlay\n"
 		" -S      start in disk usage analyser mode\n"
-		" -v      show program version and exit\n"
-		" -h      show this help and exit\n\n"
+		" -v      show program version\n"
+		" -h      show this help\n\n"
 		"Version: %s\n%s\n", VERSION, GENERAL_INFO);
 	exit(0);
 }
@@ -3340,7 +3373,7 @@ int
 main(int argc, char *argv[])
 {
 	static char cwd[PATH_MAX] __attribute__ ((aligned));
-	char *ipath = NULL, *ifilter, *bmstr;
+	char *ipath = NULL, *ifilter;
 	int opt;
 
 	/* Confirm we are in a terminal */
@@ -3353,6 +3386,7 @@ main(int argc, char *argv[])
 		switch (opt) {
 		case 'S':
 			cfg.blkorder = 1;
+			nftw_fn = sum_bsizes;
 			break;
 		case 'l':
 			cfg.showdetail = 0;
@@ -3377,7 +3411,7 @@ main(int argc, char *argv[])
 			player = optarg;
 			break;
 		case 'v':
-			printf("%s\n", VERSION);
+			fprintf(stdout, "%s\n", VERSION);
 			return 0;
 		case 'h': // fallthrough
 		default:
@@ -3385,10 +3419,8 @@ main(int argc, char *argv[])
 		}
 	}
 
-	/* Parse bookmarks string, if available */
-	bmstr = getenv("NNN_BMS");
-	if (bmstr)
-		parsebmstr(bmstr);
+	/* Parse bookmarks string */
+	parsebmstr();
 
 	if (ipath) { /* Open a bookmark directly */
 		if (get_bm_loc(ipath, cwd) == NULL) {
@@ -3430,9 +3462,6 @@ main(int argc, char *argv[])
 		fprintf(stderr, "kqueue init! %s\n", strerror(errno));
 		exit(1);
 	}
-
-	gtimeout.tv_sec = 0;
-	gtimeout.tv_nsec = 0;
 #endif
 
 	/* Edit text in EDITOR, if opted */
@@ -3447,12 +3476,14 @@ main(int argc, char *argv[])
 		player = utils[NLAY];
 
 	/* Get the desktop file manager, if set */
-	desktop_manager = getenv("NNN_DE_FILE_MANAGER");
+	dmanager = getenv("NNN_DE_FILE_MANAGER");
 
 	/* Get screensaver wait time, if set; copier used as tmp var */
 	copier = getenv("NNN_IDLE_TIMEOUT");
-	if (copier)
-		idletimeout = abs(atoi(copier));
+	if (copier) {
+		opt = atoi(copier);
+		idletimeout = opt * ((opt > 0) - (opt < 0));
+	}
 
 	/* Get the default copier, if set */
 	copier = getenv("NNN_COPIER");
@@ -3461,14 +3492,15 @@ main(int argc, char *argv[])
 	if (getenv("NNN_QUOTE_ON"))
 		cfg.quote = 1;
 
-	/* Check if X11 is available */
-	if (getenv("NNN_NO_X")) {
-		cfg.noxdisplay = 1;
+	if (getenv("HOME")) {
+		g_homelen = xstrlcpy(g_homepath, getenv("HOME"), MAX_HOME_LEN);
 
-		struct passwd *pass = getpwuid(getuid());
-
-		xstrlcpy(g_cppath, "/tmp/nnncp", 11);
-		xstrlcpy(g_cppath + 10, pass->pw_name, 33);
+		/* Check if X11 is available */
+		if (getenv("NNN_NO_X")) {
+			cfg.noxdisplay = 1;
+			xstrlcpy(g_cppath, g_homepath, MAX_HOME_LEN);
+			xstrlcpy(g_cppath + g_homelen - 1, "/.nnncp", MAX_HOME_LEN - g_homelen);
+		}
 	}
 
 	signal(SIGINT, SIG_IGN);
@@ -3490,6 +3522,9 @@ main(int argc, char *argv[])
 	browse(ipath, ifilter);
 	exitcurses();
 
+	if (cfg.noxdisplay && g_cppath[0])
+		unlink(g_cppath);
+
 #ifdef LINUX_INOTIFY
 	/* Shutdown inotify */
 	if (inotify_wd >= 0)
